#pragma once
#include "utils.h"
#include "nmflib_export.h"

extern NMFLIB_EXPORT std::unordered_map<std::string, algotype> algomap;
extern NMFLIB_EXPORT std::unordered_map<std::string, algotype> symmap;
extern NMFLIB_EXPORT std::unordered_map<std::string, normtype> normmap;
